from re import Pattern
from typing import Any


def repr_with_args(self: Any, *args, **kwargs):
    args_repr = [_smart_repr(arg) for arg in args]
    kwargs_repr = [f"{key}={_smart_repr(value)}" for key, value in kwargs.items()]
    args_and_kwargs = ", ".join(args_repr + kwargs_repr)
    return f"{self.__class__.__name__}({args_and_kwargs})"


def _smart_repr(value: Any) -> str:
    if isinstance(value, Pattern):
        # Since we're using this for codegen, we need to use our own implementation
        # of repr(...) since python truncates the pattern in its default impl
        # See: https://stackoverflow.com/a/30222089
        return f"re.compile({repr(value.pattern)})"

    return repr(value)



import json
import os.path
from os import listdir
from typing import List
from typing import Tuple

from tracking_numbers.definition import TrackingNumberDefinition
from tracking_numbers.types import Courier
from tracking_numbers.types import Spec

DEFAULT_BASE_DIR = "tracking_number_data/couriers"

TestCase = Tuple[TrackingNumberDefinition, str, bool]


def iter_courier_specs(base_dir: str = DEFAULT_BASE_DIR):
    for filename in listdir(base_dir):
        path = os.path.join(base_dir, filename)
        with open(path) as f:
            yield json.load(f)


def iter_definitions(courier_spec: Spec):
    courier = Courier(
        name=courier_spec["name"],
        code=courier_spec["courier_code"],
    )

    for tn_spec in courier_spec["tracking_numbers"]:
        definition = TrackingNumberDefinition.from_spec(courier, tn_spec)
        yield definition, tn_spec


def iter_test_cases(courier_spec: Spec) -> List[TestCase]:
    test_cases: List[TestCase] = []
    for definition, tn_spec in iter_definitions(courier_spec):
        test_numbers = tn_spec.get("test_numbers")
        if not test_numbers:
            continue

        valid_numbers = test_numbers.get("valid", [])
        for valid_number in valid_numbers:
            test_cases.append((definition, valid_number, True))

        invalid_numbers = test_numbers.get("invalid", [])
        for invalid_number in invalid_numbers:
            test_cases.append((definition, invalid_number, False))

    return test_cases



import os
from typing import List
from typing import Optional

from tracking_numbers.definition import TrackingNumberDefinition
from tracking_numbers.types import TrackingNumber

if not os.environ.get("CODE_GENERATING"):
    from tracking_numbers._generated import DEFINITIONS
else:
    # When running codegen, it's very possible that the items in
    # DEFINITIONS are out of date / can't be successfully constructed
    # so we use an empty list so that codegen can still import utils
    DEFINITIONS = []


def get_tracking_number(number: str) -> Optional[TrackingNumber]:
    for tn_definition in DEFINITIONS:
        tracking_number = tn_definition.test(number)
        if tracking_number and tracking_number.valid:
            return tracking_number

    return None


def get_definition(product_name: str) -> Optional[TrackingNumberDefinition]:
    for tn_definition in DEFINITIONS:
        if tn_definition.product.name.lower() == product_name.lower():
            return tn_definition

    return None



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

Spec = Dict[str, Any]
SerialNumber = List[int]
ValidationError = Tuple[str, str]


@dataclass
class Product:
    name: str


@dataclass
class Courier:
    code: str
    name: str


@dataclass
class TrackingNumber:
    number: str
    courier: Courier
    product: Product
    serial_number: Optional[SerialNumber]
    tracking_url: Optional[str]
    validation_errors: List[ValidationError]

    @property
    def valid(self) -> bool:
        return not self.validation_errors


def to_int(serial_number: SerialNumber) -> int:
    return int("".join(map(str, serial_number)))



import re
from abc import ABCMeta
from abc import abstractmethod
from re import Pattern

from tracking_numbers.helpers.repr import repr_with_args
from tracking_numbers.types import Spec


class ValueMatcher(metaclass=ABCMeta):
    @abstractmethod
    def __repr__(self):
        raise NotImplementedError

    @abstractmethod
    def matches(self, other: str) -> bool:
        raise NotImplementedError

    @classmethod
    def from_spec(cls, spec: Spec) -> "ValueMatcher":
        if "matches" in spec:
            return ExactValueMatcher(spec["matches"])
        elif "matches_regex" in spec:
            return RegexValueMatcher(spec["matches_regex"])

        raise ValueError(f"Invalid matcher spec: {spec}")


class ExactValueMatcher(ValueMatcher):
    def __init__(self, value: str):
        self.value = value

    def __repr__(self):
        return repr_with_args(self, value=self.value)

    def matches(self, other: str) -> bool:
        return self.value == other


class RegexValueMatcher(ValueMatcher):
    def __init__(self, pattern: Pattern):
        self.pattern = re.compile(pattern)

    def __repr__(self):
        return repr_with_args(self, pattern=self.pattern)

    def matches(self, other: str) -> bool:
        return bool(self.pattern.match(other))



from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import Optional

from tracking_numbers.helpers.repr import repr_with_args
from tracking_numbers.types import SerialNumber
from tracking_numbers.types import Spec
from tracking_numbers.types import to_int


class ChecksumValidator(metaclass=ABCMeta):
    def __repr__(self):
        return repr_with_args(self)

    @abstractmethod
    def passes(self, serial_number: SerialNumber, check_digit: int) -> bool:
        raise NotImplementedError

    @classmethod
    def from_spec(cls, validation_spec: Spec) -> Optional["ChecksumValidator"]:
        checksum_spec = validation_spec.get("checksum")
        if not checksum_spec:
            return None

        strategy = checksum_spec.get("name")
        if strategy == "s10":
            return S10()
        elif strategy == "mod7":
            return Mod7()
        elif strategy == "mod10":
            return Mod10(
                odds_multiplier=checksum_spec.get("odds_multiplier"),
                evens_multiplier=checksum_spec.get("evens_multiplier"),
            )
        elif strategy == "sum_product_with_weightings_and_modulo":
            return SumProductWithWeightsAndModulo(
                weights=checksum_spec["weightings"],
                first_modulo=checksum_spec["modulo1"],
                second_modulo=checksum_spec["modulo2"],
            )

        raise ValueError(f"Unknown checksum: {strategy}")


class S10(ChecksumValidator):
    WEIGHTS = [8, 6, 4, 2, 3, 5, 9, 7]

    def passes(self, serial_number: SerialNumber, check_digit: int) -> bool:
        total = 0
        for digit, weight in zip(serial_number, self.WEIGHTS):
            total += digit * weight

        remainder = total % 11
        if remainder == 1:
            check = 0
        elif remainder == 0:
            check = 5
        else:
            check = 11 - remainder

        return check == check_digit


class Mod10(ChecksumValidator):
    def __init__(
        self,
        odds_multiplier: Optional[int] = None,
        evens_multiplier: Optional[int] = None,
    ):
        self.odds_multiplier = odds_multiplier
        self.evens_multiplier = evens_multiplier

    def __repr__(self):
        return repr_with_args(
            self,
            odds_multiplier=self.odds_multiplier,
            evens_multiplier=self.evens_multiplier,
        )

    def passes(self, serial_number: SerialNumber, check_digit: int) -> bool:
        total = 0
        for index, digit in enumerate(serial_number):
            is_even_index = index % 2 == 0
            is_odd_index = not is_even_index

            if is_odd_index and self.odds_multiplier:
                total += digit * self.odds_multiplier
            elif is_even_index and self.evens_multiplier:
                total += digit * self.evens_multiplier
            else:
                total += digit

        check = total % 10
        if check != 0:
            check = 10 - check

        return check == check_digit


class Mod7(ChecksumValidator):
    def passes(self, serial_number: SerialNumber, check_digit: int) -> bool:
        return check_digit == (to_int(serial_number) % 7)


class SumProductWithWeightsAndModulo(ChecksumValidator):
    def __init__(self, weights: List[int], first_modulo: int, second_modulo: int):
        self.weights = weights
        self.first_modulo = first_modulo
        self.second_modulo = second_modulo

    def __repr__(self):
        return repr_with_args(
            self,
            weights=self.weights,
            first_modulo=self.first_modulo,
            second_modulo=self.second_modulo,
        )

    def passes(self, serial_number: SerialNumber, check_digit: int) -> bool:
        total = 0
        for digit, weight in zip(serial_number, self.weights):
            total += digit * weight

        check = total % self.first_modulo % self.second_modulo
        return check == check_digit



import re
from re import Pattern
from typing import List
from typing import Union


def pcre_to_python_re(regex: str) -> Pattern:
    """Converts a PCRE (Perl) to a Python-compatible regex"""
    return re.compile(regex.replace("(?<", "(?P<"))


def parse_regex(raw_regex: Union[str, List[str]]) -> Pattern:
    """Parses a single- or multi-line regex, where lines are represented
    as items in a list.
    """
    if isinstance(raw_regex, list):
        raw_regex = "".join(raw_regex)

    return pcre_to_python_re(raw_regex)



from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from re import Pattern
from typing import Optional

from tracking_numbers.compat import pcre_to_python_re
from tracking_numbers.helpers.repr import repr_with_args
from tracking_numbers.types import SerialNumber
from tracking_numbers.types import Spec


@dataclass
class PrependIf:
    matches_regex: Pattern
    content: str

    def apply(self, serial_number: str) -> str:
        return (
            self.content + serial_number
            if self.matches_regex.match(serial_number)
            else serial_number
        )


class SerialNumberParser(metaclass=ABCMeta):
    def __repr__(self):
        return repr_with_args(self)

    @abstractmethod
    def parse(self, number: str) -> SerialNumber:
        raise NotImplementedError


class DefaultSerialNumberParser(SerialNumberParser):
    def __init__(self, prepend_if: Optional[PrependIf] = None):
        self.prepend_if = prepend_if

    def __repr__(self):
        return repr_with_args(self, prepend_if=self.prepend_if)

    def parse(self, number: str) -> SerialNumber:
        if self.prepend_if:
            number = self.prepend_if.apply(number)

        return [int(digit) for digit in number]

    @classmethod
    def from_spec(cls, validation_spec: Spec) -> "SerialNumberParser":
        serial_number_format = validation_spec.get("serial_number_format")
        if not serial_number_format:
            return DefaultSerialNumberParser()

        prepend_if_spec = serial_number_format.get("prepend_if")
        if not prepend_if_spec:
            return DefaultSerialNumberParser()

        prepend_if = PrependIf(
            matches_regex=pcre_to_python_re(prepend_if_spec["matches_regex"]),
            content=prepend_if_spec["content"],
        )

        return DefaultSerialNumberParser(prepend_if)


class UPSSerialNumberParser(SerialNumberParser):
    def __repr__(self):
        return repr_with_args(self)

    def parse(self, number: str) -> SerialNumber:
        return [self._value_of(ch) for ch in number]

    @staticmethod
    def _value_of(ch: str) -> int:
        # Can't find a definitive spec for _why_ the chars are mapped this way,
        # but I did manage to find the following articles that help to confirm
        # https://abelable.altervista.org/check-digit-function-for-an-ups-tracking-number/
        # https://www.codeproject.com/articles/21224/calculating-the-ups-tracking-number-check-digit
        return int(ch) if ch.isdigit() else (ord(ch) - 3) % 10



# DO NOT EDIT - Generated by codegen.py
import re

from tracking_numbers.checksum_validator import Mod10
from tracking_numbers.checksum_validator import Mod7
from tracking_numbers.checksum_validator import S10
from tracking_numbers.checksum_validator import SumProductWithWeightsAndModulo
from tracking_numbers.definition import AdditionalValidation
from tracking_numbers.definition import TrackingNumberDefinition
from tracking_numbers.serial_number import DefaultSerialNumberParser
from tracking_numbers.serial_number import PrependIf
from tracking_numbers.serial_number import UPSSerialNumberParser
from tracking_numbers.types import Courier
from tracking_numbers.types import Product
from tracking_numbers.value_matcher import ExactValueMatcher
from tracking_numbers.value_matcher import RegexValueMatcher


DEFINITIONS = [
    TrackingNumberDefinition(
        courier=Courier(code="cdl", name="CDL"),
        product=Product(name="CDL Last Mile Solutions"),
        number_regex=re.compile(
            "\\s*(?=.*[a-z])(?P<PackageId>([0-9a-f]\\s*){10,10})\\s*",
        ),
        tracking_url_template="https://ship.cdldelivers.com/Xcelerator/Tracking/Tracking?packageitemrefno=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=None,
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="dhl", name="DHL"),
        product=Product(name="DHL Express"),
        number_regex=re.compile(
            "\\s*(?P<SerialNumber>([0-9]\\s*){9})(?P<CheckDigit>([0-9]\\s*))",
        ),
        tracking_url_template="http://www.dhl.com/en/express/tracking.html?brand=DHL&AWB=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod7(),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="dhl", name="DHL"),
        product=Product(name="DHL Express Air"),
        number_regex=re.compile(
            "\\s*(?P<SerialNumber>([0-9]\\s*){10})(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="http://www.dhl.com/en/express/tracking.html?brand=DHL&AWB=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod7(),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="amazon", name="Amazon"),
        product=Product(name="Amazon Logistics"),
        number_regex=re.compile(
            "\\s*T\\s*B\\s*A\\s*(?P<SerialNumber>([0-9]\\s*){12,12})\\s*",
        ),
        tracking_url_template=None,
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=None,
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="usps", name="United States Postal Service"),
        product=Product(name="USPS 20"),
        number_regex=re.compile(
            "\\s*(?P<SerialNumber>(?P<ServiceType>([0-9]\\s*){2})(?P<ShipperId>([0-9]\\s*){9})(?P<PackageId>([0-9]\\s*){8}))(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://tools.usps.com/go/TrackConfirmAction?tLabels=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod10(odds_multiplier=1, evens_multiplier=3),
        additional_validations=[
            AdditionalValidation(
                name="Service Type",
                regex_group_name="ServiceType",
                value_matchers=[
                    ExactValueMatcher(value="03"),
                    ExactValueMatcher(value="71"),
                    ExactValueMatcher(value="73"),
                    ExactValueMatcher(value="77"),
                    ExactValueMatcher(value="81"),
                ],
            ),
        ],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="usps", name="United States Postal Service"),
        product=Product(name="USPS 34v2"),
        number_regex=re.compile(
            "\\s*(?P<RoutingApplicationId>4\\s*2\\s*0\\s*)(?P<DestinationZip>([0-9]\\s*){5})(?P<RoutingNumber>([0-9]\\s*){4})(?P<SerialNumber>(?P<ApplicationIdentifier>9\\s*[2345]\\s*)?(?P<ShipperId>([0-9]\\s*){8})(?P<PackageId>([0-9]\\s*){11}))(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://tools.usps.com/go/TrackConfirmAction?tLabels=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod10(odds_multiplier=1, evens_multiplier=3),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="usps", name="United States Postal Service"),
        product=Product(name="USPS 91"),
        number_regex=re.compile(
            "\\s*(?:(?P<RoutingApplicationId>4\\s*2\\s*0\\s*)(?P<DestinationZip>([0-9]\\s*){5}))?(?P<SerialNumber>(?P<ApplicationIdentifier>9\\s*[12345]\\s*)?(?P<SCNC>([0-9]\\s*){2})(?P<ServiceType>([0-9]\\s*){2})(?P<ShipperId>([0-9]\\s*){8})(?P<PackageId>([0-9]\\s*){11}|([0-9]\\s*){7}))(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://tools.usps.com/go/TrackConfirmAction?tLabels=%s",
        serial_number_parser=DefaultSerialNumberParser(
            prepend_if=PrependIf(
                matches_regex=re.compile("^(?!9[1-5]).+"),
                content="91",
            ),
        ),
        checksum_validator=Mod10(odds_multiplier=1, evens_multiplier=3),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx Express (12)"),
        number_regex=re.compile(
            "\\s*(?P<SerialNumber>([0-9]\\s*){11})(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=SumProductWithWeightsAndModulo(
            weights=[3, 1, 7, 3, 1, 7, 3, 1, 7, 3, 1],
            first_modulo=11,
            second_modulo=10,
        ),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx Express (34)"),
        number_regex=re.compile(
            "\\s*1\\s*0\\s*0\\s*[0-9]\\s*[0-9]\\s*([0-9]\\s*){10}(?P<DestinationZip>([0-9]\\s*){5})(?P<SerialNumber>([0-9]\\s*){13})(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=SumProductWithWeightsAndModulo(
            weights=[1, 7, 3, 1, 7, 3, 1, 7, 3, 1, 7, 3, 1],
            first_modulo=11,
            second_modulo=10,
        ),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx SmartPost"),
        number_regex=re.compile(
            "\\s*(?P<ApplicationIdentifier>9\\s*2\\s*)?(?P<SerialNumber>(?P<ServiceType>([0-9]\\s*){3})(?P<ShipperId>([0-9]\\s*){9})(?P<PackageId>([0-9]\\s*){7}))(?P<CheckDigit>([0-9]\\s*))",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(
            prepend_if=PrependIf(matches_regex=re.compile("^(?!92).+"), content="92"),
        ),
        checksum_validator=Mod10(odds_multiplier=1, evens_multiplier=3),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx Ground"),
        number_regex=re.compile(
            "\\s*(?P<SerialNumber>([0-9]\\s*){14})(?P<CheckDigit>([0-9]\\s*))",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod10(odds_multiplier=3, evens_multiplier=1),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx Ground (SSCC-18)"),
        number_regex=re.compile(
            "\\s*(?P<ShippingContainerType>([0-9]\\s*){2})(?P<SerialNumber>([0-9]\\s*){15})(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod10(odds_multiplier=1, evens_multiplier=3),
        additional_validations=[
            AdditionalValidation(
                name="Container Type",
                regex_group_name="ShippingContainerType",
                value_matchers=[
                    ExactValueMatcher(value="00"),
                    ExactValueMatcher(value="01"),
                    ExactValueMatcher(value="02"),
                    ExactValueMatcher(value="04"),
                ],
            ),
        ],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx Ground 96 (22)"),
        number_regex=re.compile(
            "\\s*(?P<ApplicationIdentifier>9\\s*6\\s*)(?P<SCNC>([0-9]\\s*){2})(?P<ServiceType>([0-9]\\s*){3})(?P<SerialNumber>(?P<ShipperId>([0-9]\\s*){7})(?P<PackageId>([0-9]\\s*){7}))(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=Mod10(odds_multiplier=3, evens_multiplier=1),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="fedex", name="FedEx"),
        product=Product(name="FedEx Ground GSN"),
        number_regex=re.compile(
            "\\s*(?P<ApplicationIdentifier>9\\s*6\\s*)(?P<SCNC>([0-9]\\s*){2})([0-9]\\s*){5}(?P<GSN>([0-9]\\s*){10})[0-9]\\s*(?P<SerialNumber>([0-9]\\s*){13})(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="https://www.fedex.com/apps/fedextrack/?tracknumbers=%s",
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=SumProductWithWeightsAndModulo(
            weights=[1, 7, 3, 1, 7, 3, 1, 7, 3, 1, 7, 3, 1],
            first_modulo=11,
            second_modulo=10,
        ),
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="ups", name="UPS"),
        product=Product(name="UPS"),
        number_regex=re.compile(
            "\\s*1\\s*Z\\s*(?P<SerialNumber>(?P<ShipperId>(?:[A-Z0-9]\\s*){6,6})(?P<ServiceType>(?:[A-Z0-9]\\s*){2,2})(?P<PackageId>(?:[A-Z0-9]\\s*){7,7}))(?P<CheckDigit>[A-Z0-9]\\s*)",
        ),
        tracking_url_template="https://wwwapps.ups.com/WebTracking/track?track=yes&trackNums=%s",
        serial_number_parser=UPSSerialNumberParser(),
        checksum_validator=Mod10(odds_multiplier=2, evens_multiplier=1),
        additional_validations=[
            AdditionalValidation(
                name="Service Type",
                regex_group_name="ServiceType",
                value_matchers=[
                    ExactValueMatcher(value="01"),
                    ExactValueMatcher(value="02"),
                    ExactValueMatcher(value="03"),
                    ExactValueMatcher(value="04"),
                    ExactValueMatcher(value="12"),
                    ExactValueMatcher(value="13"),
                    ExactValueMatcher(value="15"),
                    ExactValueMatcher(value="22"),
                    ExactValueMatcher(value="32"),
                    ExactValueMatcher(value="33"),
                    ExactValueMatcher(value="41"),
                    ExactValueMatcher(value="42"),
                    ExactValueMatcher(value="44"),
                    ExactValueMatcher(value="66"),
                    ExactValueMatcher(value="67"),
                    ExactValueMatcher(value="72"),
                    ExactValueMatcher(value="78"),
                    ExactValueMatcher(value="90"),
                    ExactValueMatcher(value="A0"),
                    ExactValueMatcher(value="A1"),
                    ExactValueMatcher(value="A2"),
                    ExactValueMatcher(value="A8"),
                    ExactValueMatcher(value="A9"),
                    ExactValueMatcher(value="AA"),
                    ExactValueMatcher(value="YW"),
                ],
            ),
        ],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="ups", name="UPS"),
        product=Product(name="UPS Mail Innovations - Sequence Number"),
        number_regex=re.compile(
            "\\s*8\\s*0\\s*(?P<SerialNumber>([0-9]\\s*){16,16})\\s*",
        ),
        tracking_url_template="https://wwwapps.ups.com/WebTracking/track?track=yes&trackNums=%s",
        serial_number_parser=UPSSerialNumberParser(),
        checksum_validator=None,
        additional_validations=[],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="s10", name="S10 International Standard"),
        product=Product(name="S10"),
        number_regex=re.compile(
            "\\s*(?P<ServiceType>([A-Z]\\s*){2})(?P<SerialNumber>([0-9]\\s*){8})(?P<CheckDigit>([0-9]\\s*))(?P<CountryCode>([A-Z]\\s*){2})",
        ),
        tracking_url_template=None,
        serial_number_parser=DefaultSerialNumberParser(prepend_if=None),
        checksum_validator=S10(),
        additional_validations=[
            AdditionalValidation(
                name="Service Type",
                regex_group_name="ServiceType",
                value_matchers=[
                    RegexValueMatcher(pattern=re.compile("E[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("L[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("M[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("Q[A-M]")),
                    RegexValueMatcher(pattern=re.compile("R[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("U[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("V[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("C[A-Z]")),
                    RegexValueMatcher(pattern=re.compile("H[A-Z]")),
                    RegexValueMatcher(
                        pattern=re.compile("([BDNPZ][A-Z]|A[V-Z]|G[AD])"),
                    ),
                ],
            ),
            AdditionalValidation(
                name="Courier",
                regex_group_name="CountryCode",
                value_matchers=[
                    ExactValueMatcher(value="AF"),
                    ExactValueMatcher(value="AL"),
                    ExactValueMatcher(value="DZ"),
                    ExactValueMatcher(value="AO"),
                    ExactValueMatcher(value="AG"),
                    ExactValueMatcher(value="AR"),
                    ExactValueMatcher(value="AM"),
                    ExactValueMatcher(value="AU"),
                    ExactValueMatcher(value="AT"),
                    ExactValueMatcher(value="AZ"),
                    ExactValueMatcher(value="BS"),
                    ExactValueMatcher(value="BH"),
                    ExactValueMatcher(value="BD"),
                    ExactValueMatcher(value="BB"),
                    ExactValueMatcher(value="BY"),
                    ExactValueMatcher(value="BE"),
                    ExactValueMatcher(value="BZ"),
                    ExactValueMatcher(value="BJ"),
                    ExactValueMatcher(value="BT"),
                    ExactValueMatcher(value="BO"),
                    ExactValueMatcher(value="BA"),
                    ExactValueMatcher(value="BW"),
                    ExactValueMatcher(value="BR"),
                    ExactValueMatcher(value="BN"),
                    ExactValueMatcher(value="BG"),
                    ExactValueMatcher(value="BF"),
                    ExactValueMatcher(value="BI"),
                    ExactValueMatcher(value="KH"),
                    ExactValueMatcher(value="CM"),
                    ExactValueMatcher(value="CA"),
                    ExactValueMatcher(value="CV"),
                    ExactValueMatcher(value="CF"),
                    ExactValueMatcher(value="TD"),
                    ExactValueMatcher(value="CL"),
                    ExactValueMatcher(value="CN"),
                    ExactValueMatcher(value="HK"),
                    ExactValueMatcher(value="CO"),
                    ExactValueMatcher(value="KM"),
                    ExactValueMatcher(value="CG"),
                    ExactValueMatcher(value="CR"),
                    ExactValueMatcher(value="HR"),
                    ExactValueMatcher(value="CU"),
                    ExactValueMatcher(value="CY"),
                    ExactValueMatcher(value="CZ"),
                    ExactValueMatcher(value="CI"),
                    ExactValueMatcher(value="KP"),
                    ExactValueMatcher(value="CD"),
                    ExactValueMatcher(value="DK"),
                    ExactValueMatcher(value="DJ"),
                    ExactValueMatcher(value="DM"),
                    ExactValueMatcher(value="DO"),
                    ExactValueMatcher(value="EC"),
                    ExactValueMatcher(value="EG"),
                    ExactValueMatcher(value="SV"),
                    ExactValueMatcher(value="GQ"),
                    ExactValueMatcher(value="ER"),
                    ExactValueMatcher(value="EE"),
                    ExactValueMatcher(value="ET"),
                    ExactValueMatcher(value="FJ"),
                    ExactValueMatcher(value="FI"),
                    ExactValueMatcher(value="FR"),
                    ExactValueMatcher(value="GA"),
                    ExactValueMatcher(value="GM"),
                    ExactValueMatcher(value="GE"),
                    ExactValueMatcher(value="DE"),
                    ExactValueMatcher(value="GH"),
                    ExactValueMatcher(value="GB"),
                    ExactValueMatcher(value="GR"),
                    ExactValueMatcher(value="GD"),
                    ExactValueMatcher(value="GT"),
                    ExactValueMatcher(value="GN"),
                    ExactValueMatcher(value="GW"),
                    ExactValueMatcher(value="GY"),
                    ExactValueMatcher(value="HT"),
                    ExactValueMatcher(value="HN"),
                    ExactValueMatcher(value="HU"),
                    ExactValueMatcher(value="IS"),
                    ExactValueMatcher(value="IN"),
                    ExactValueMatcher(value="ID"),
                    ExactValueMatcher(value="IR"),
                    ExactValueMatcher(value="IQ"),
                    ExactValueMatcher(value="IE"),
                    ExactValueMatcher(value="IL"),
                    ExactValueMatcher(value="IT"),
                    ExactValueMatcher(value="JM"),
                    ExactValueMatcher(value="JP"),
                    ExactValueMatcher(value="JO"),
                    ExactValueMatcher(value="KZ"),
                    ExactValueMatcher(value="KE"),
                    ExactValueMatcher(value="KI"),
                    ExactValueMatcher(value="KR"),
                    ExactValueMatcher(value="KW"),
                    ExactValueMatcher(value="KG"),
                    ExactValueMatcher(value="LA"),
                    ExactValueMatcher(value="LV"),
                    ExactValueMatcher(value="LB"),
                    ExactValueMatcher(value="LS"),
                    ExactValueMatcher(value="LR"),
                    ExactValueMatcher(value="LY"),
                    ExactValueMatcher(value="LI"),
                    ExactValueMatcher(value="LT"),
                    ExactValueMatcher(value="LU"),
                    ExactValueMatcher(value="MG"),
                    ExactValueMatcher(value="MW"),
                    ExactValueMatcher(value="MY"),
                    ExactValueMatcher(value="MV"),
                    ExactValueMatcher(value="ML"),
                    ExactValueMatcher(value="MT"),
                    ExactValueMatcher(value="MR"),
                    ExactValueMatcher(value="MU"),
                    ExactValueMatcher(value="MX"),
                    ExactValueMatcher(value="MD"),
                    ExactValueMatcher(value="MC"),
                    ExactValueMatcher(value="MN"),
                    ExactValueMatcher(value="ME"),
                    ExactValueMatcher(value="MA"),
                    ExactValueMatcher(value="MZ"),
                    ExactValueMatcher(value="MM"),
                    ExactValueMatcher(value="NA"),
                    ExactValueMatcher(value="NR"),
                    ExactValueMatcher(value="NP"),
                    ExactValueMatcher(value="NL"),
                    ExactValueMatcher(value="NZ"),
                    ExactValueMatcher(value="NI"),
                    ExactValueMatcher(value="NE"),
                    ExactValueMatcher(value="NG"),
                    ExactValueMatcher(value="NO"),
                    ExactValueMatcher(value="OM"),
                    ExactValueMatcher(value="PK"),
                    ExactValueMatcher(value="PA"),
                    ExactValueMatcher(value="PG"),
                    ExactValueMatcher(value="PY"),
                    ExactValueMatcher(value="PE"),
                    ExactValueMatcher(value="PH"),
                    ExactValueMatcher(value="PL"),
                    ExactValueMatcher(value="PT"),
                    ExactValueMatcher(value="QA"),
                    ExactValueMatcher(value="RO"),
                    ExactValueMatcher(value="RU"),
                    ExactValueMatcher(value="RW"),
                    ExactValueMatcher(value="KN"),
                    ExactValueMatcher(value="LC"),
                    ExactValueMatcher(value="VC"),
                    ExactValueMatcher(value="WS"),
                    ExactValueMatcher(value="SM"),
                    ExactValueMatcher(value="ST"),
                    ExactValueMatcher(value="SA"),
                    ExactValueMatcher(value="SN"),
                    ExactValueMatcher(value="RS"),
                    ExactValueMatcher(value="SC"),
                    ExactValueMatcher(value="SL"),
                    ExactValueMatcher(value="SG"),
                    ExactValueMatcher(value="SK"),
                    ExactValueMatcher(value="SI"),
                    ExactValueMatcher(value="SB"),
                    ExactValueMatcher(value="SO"),
                    ExactValueMatcher(value="ZA"),
                    ExactValueMatcher(value="SS"),
                    ExactValueMatcher(value="ES"),
                    ExactValueMatcher(value="LK"),
                    ExactValueMatcher(value="SD"),
                    ExactValueMatcher(value="SR"),
                    ExactValueMatcher(value="SZ"),
                    ExactValueMatcher(value="SE"),
                    ExactValueMatcher(value="CH"),
                    ExactValueMatcher(value="SY"),
                    ExactValueMatcher(value="TJ"),
                    ExactValueMatcher(value="TZ"),
                    ExactValueMatcher(value="TH"),
                    ExactValueMatcher(value="MK"),
                    ExactValueMatcher(value="TL"),
                    ExactValueMatcher(value="TG"),
                    ExactValueMatcher(value="TO"),
                    ExactValueMatcher(value="TT"),
                    ExactValueMatcher(value="TN"),
                    ExactValueMatcher(value="TR"),
                    ExactValueMatcher(value="TM"),
                    ExactValueMatcher(value="TV"),
                    ExactValueMatcher(value="UG"),
                    ExactValueMatcher(value="UA"),
                    ExactValueMatcher(value="AE"),
                    ExactValueMatcher(value="US"),
                    ExactValueMatcher(value="UY"),
                    ExactValueMatcher(value="UZ"),
                    ExactValueMatcher(value="VU"),
                    ExactValueMatcher(value="VA"),
                    ExactValueMatcher(value="VE"),
                    ExactValueMatcher(value="VN"),
                    ExactValueMatcher(value="YE"),
                    ExactValueMatcher(value="ZM"),
                    ExactValueMatcher(value="ZW"),
                ],
            ),
        ],
    ),
    TrackingNumberDefinition(
        courier=Courier(code="ontrac", name="OnTrac"),
        product=Product(name="OnTrac"),
        number_regex=re.compile(
            "\\s*C\\s*(?P<SerialNumber>([0-9]\\s*){13})(?P<CheckDigit>[0-9]\\s*)",
        ),
        tracking_url_template="http://www.ontrac.com/trackingres.asp?tracking_number=%s",
        serial_number_parser=DefaultSerialNumberParser(
            prepend_if=PrependIf(matches_regex=re.compile("^(?!4).+$"), content="4"),
        ),
        checksum_validator=Mod10(odds_multiplier=2, evens_multiplier=1),
        additional_validations=[],
    ),
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tracking_numbers', 'tracking_numbers.helpers']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'tracking-numbers',
    'version': '0.1.6',
    'description': 'Parse tracking numbers from couriers with no external dependencies',
    'long_description': '<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n<!-- DON\'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->\n**Table of Contents**  *generated with [DocToc](https://github.com/thlorenz/doctoc)*\n\n- [tracking-numbers](#tracking-numbers)\n  - [Why?](#why)\n  - [Installation](#installation)\n  - [Usage](#usage)\n    - [`get_tracking_number(number)`](#get_tracking_numbernumber)\n    - [`get_definition(product_name)`](#get_definitionproduct_name)\n  - [Testing](#testing)\n\n<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n\n# tracking-numbers\n\nA library that parses tracking numbers and provides common types.\nThe data is sourced from [`tracking_number_data`](https://github.com/jkeen/tracking_number_data/) and the definitions are code-generated.\n\n## Why?\n\nThe typical shipping tracking number has a lot of data encoded within.\nWhile some couriers share similar logic (serial number, check digit, etc), others have entirely different ways of representing tracking numbers.\n\nInstead of hand-rolling parsing code for all of these cases, the author of [`tracking_number_data`](https://github.com/jkeen/tracking_number_data/) has put together a repo that serves as a language-agnostic source of knowledge for various couriers and their shipping products.\n\nThis library uses that data to code-generate definitions to create python bindings for parsing tracking numbers.\n\nThe library itself has no external dependencies, and can be used to decode basic tracking data without the need of an API or external data source at runtime.\n\n## Installation\n\nInstall the tracking-numbers library using pypi\n\n```sh\npip install tracking-numbers\n```\n\n## Usage\n\nHere are the main public functions to use:\n\n### `get_tracking_number(number)`\n\nParses the `number` and returns the `TrackingNumber` dataclass, if detected, or none otherwise.\n\n```python\nfrom tracking_numbers import get_tracking_number\n\ntracking_number = get_tracking_number("1ZY0X1930320121606")\n\n# => TrackingNumber(\n#       valid=False,\n#       number=\'1ZY0X1930320121606\',\n#       serial_number=[6, 0, 5, 1, 9, 3, 0, 3, 2, 0, 1, 2, 1, 6, 0],\n#       tracking_url=\'https://wwwapps.ups.com/WebTracking/track?track=yes&trackNums=1ZY0X1930320121604\',\n#       courier=Courier(code=\'ups\', name=\'UPS\'),\n#       product=Product(name=\'UPS\'),\n#    )\n```\n\n### `get_definition(product_name)`\n\nGiven a product name, gets the `TrackingNumberDefinition` associated.\nYou can call `definition.test(number)` to parse a number for that specific product.\n\n```python\nfrom tracking_numbers import get_definition\n\nups_definition = get_definition(\'UPS\')\ntracking_number = ups_definition.test("1ZY0X1930320121606")\n\n# => TrackingNumber(\n#       valid=False,\n#       number=\'1ZY0X1930320121606\',\n#       serial_number=[6, 0, 5, 1, 9, 3, 0, 3, 2, 0, 1, 2, 1, 6, 0],\n#       tracking_url=\'https://wwwapps.ups.com/WebTracking/track?track=yes&trackNums=1ZY0X1930320121604\',\n#       courier=Courier(code=\'ups\', name=\'UPS\'),\n#       product=Product(name=\'UPS\'),\n#    )\n\ntracking_number = ups_definition.test(\'some_valid_fedex_number\')\n\n# => None\n```\n\n## Testing\n\nWe use the test cases defined in the courier data to generate pytest test cases.\nIn this way, we can be confident that the logic for parsing tracking numbers is working properly.\n',
    'author': 'Jonathan Como',
    'author_email': 'public@jcomo.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jcomo/tracking-numbers',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Pattern

from tracking_numbers.checksum_validator import ChecksumValidator
from tracking_numbers.compat import parse_regex
from tracking_numbers.helpers.repr import repr_with_args
from tracking_numbers.serial_number import DefaultSerialNumberParser
from tracking_numbers.serial_number import SerialNumberParser
from tracking_numbers.serial_number import UPSSerialNumberParser
from tracking_numbers.types import Courier
from tracking_numbers.types import Product
from tracking_numbers.types import SerialNumber
from tracking_numbers.types import Spec
from tracking_numbers.types import TrackingNumber
from tracking_numbers.types import ValidationError
from tracking_numbers.value_matcher import ValueMatcher

MatchData = Dict[str, str]


@dataclass
class AdditionalValidation:
    name: str
    regex_group_name: str
    value_matchers: List[ValueMatcher]

    @classmethod
    def from_spec(cls, spec: Spec) -> "AdditionalValidation":
        value_matchers: List[ValueMatcher] = []
        for value_matcher_spec in spec["lookup"]:
            value_matchers.append(ValueMatcher.from_spec(value_matcher_spec))

        return AdditionalValidation(
            name=spec["name"],
            regex_group_name=spec["regex_group_name"],
            value_matchers=value_matchers,
        )


class TrackingNumberDefinition:
    courier: Courier
    product: Product
    number_regex: Pattern
    tracking_url_template: Optional[str]
    serial_number_parser: SerialNumberParser
    checksum_validator: Optional[ChecksumValidator]
    additional_validations: List[AdditionalValidation]

    def __init__(
        self,
        courier: Courier,
        product: Product,
        number_regex: Pattern,
        tracking_url_template: Optional[str],
        serial_number_parser: SerialNumberParser,
        checksum_validator: Optional[ChecksumValidator],
        additional_validations: List[AdditionalValidation],
    ):
        self.courier = courier
        self.product = product
        self.number_regex = number_regex
        self.tracking_url_template = tracking_url_template
        self.serial_number_parser = serial_number_parser
        self.checksum_validator = checksum_validator
        self.additional_validations = additional_validations

    def __repr__(self):
        return repr_with_args(
            self,
            courier=self.courier,
            product=self.product,
            number_regex=self.number_regex,
            tracking_url_template=self.tracking_url_template,
            serial_number_parser=self.serial_number_parser,
            checksum_validator=self.checksum_validator,
            additional_validations=self.additional_validations,
        )

    @classmethod
    def from_spec(cls, courier: Courier, tn_spec: Spec) -> "TrackingNumberDefinition":
        product = Product(name=tn_spec["name"])
        tracking_url_template = tn_spec.get("tracking_url")
        number_regex = parse_regex(tn_spec["regex"])

        validation_spec = tn_spec["validation"]
        serial_number_parser = (
            UPSSerialNumberParser()
            if courier.code == "ups"
            else DefaultSerialNumberParser.from_spec(validation_spec)
        )

        additional_spec = tn_spec.get("additional")
        additional_validations: List[AdditionalValidation] = []
        if isinstance(additional_spec, list):
            # Handles None and 1 that is a dict (seems like old format / mistake)
            for spec in additional_spec:
                additional_validations.append(AdditionalValidation.from_spec(spec))

        return TrackingNumberDefinition(
            courier=courier,
            product=product,
            number_regex=number_regex,
            tracking_url_template=tracking_url_template,
            serial_number_parser=serial_number_parser,
            checksum_validator=ChecksumValidator.from_spec(validation_spec),
            additional_validations=additional_validations,
        )

    def test(self, tracking_number: str) -> Optional[TrackingNumber]:
        match = self.number_regex.fullmatch(tracking_number)
        if not match:
            return None

        match_data = match.groupdict() if match else {}
        serial_number = self._get_serial_number(match_data)
        validation_errors = self._get_validation_errors(serial_number, match_data)

        return TrackingNumber(
            number=tracking_number,
            courier=self.courier,
            product=self.product,
            serial_number=serial_number,
            tracking_url=self.tracking_url(tracking_number),
            validation_errors=validation_errors,
        )

    def _get_serial_number(self, match_data: MatchData) -> Optional[SerialNumber]:
        raw_serial_number = match_data.get("SerialNumber")
        if raw_serial_number:
            return self.serial_number_parser.parse(
                _remove_whitespace(raw_serial_number),
            )

        return None

    def _get_validation_errors(
        self,
        serial_number: Optional[SerialNumber],
        match_data: MatchData,
    ) -> List[ValidationError]:
        errors: List[ValidationError] = []
        checksum_error = self._get_checksum_errors(serial_number, match_data)
        if checksum_error:
            errors.append(checksum_error)

        for validation in self.additional_validations:
            additional_error = self._get_additional_error(validation, match_data)
            if additional_error:
                errors.append(additional_error)

        return errors

    def _get_checksum_errors(
        self,
        serial_number: Optional[SerialNumber],
        match_data: MatchData,
    ) -> Optional[ValidationError]:
        if not self.checksum_validator:
            return None

        if not serial_number:
            return "checksum", "SerialNumber not found"

        check_digit = match_data.get("CheckDigit")
        if not check_digit:
            return "checksum", "CheckDigit not found"

        passes_checksum = self.checksum_validator.passes(
            serial_number=serial_number,
            check_digit=int(check_digit),
        )

        if not passes_checksum:
            return "checksum", "Checksum validation failed"

        return None

    @staticmethod
    def _get_additional_error(
        validation: AdditionalValidation,
        match_data: MatchData,
    ) -> Optional[ValidationError]:
        group_key = validation.regex_group_name
        raw_value = match_data.get(group_key)
        if not raw_value:
            return validation.name, f"{group_key} not found"

        value = _remove_whitespace(raw_value)
        matches_any_value = any(
            value_matcher.matches(value) for value_matcher in validation.value_matchers
        )

        if not matches_any_value:
            return validation.name, f"Match not found for {group_key}: {value}"

        return None

    def tracking_url(self, tracking_number: str) -> Optional[str]:
        if not self.tracking_url_template:
            return None

        return self.tracking_url_template % tracking_number


def _remove_whitespace(value: str) -> str:
    return "".join(ch for ch in value if ch.strip())
